/*
 * MKLOADOB.C
 *
 *  COPYRIGHT : 1993 Echidna.
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 05/09/93
 *   MODIFIED : 09/08/94
 *       TABS : 05 09
 *
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Takes a bunch of binary files and concatinates them, checks for
 *		duplicates and creates a 2 word block relative offset table for each
 *		piece of data may cross a 2048 byte boundry.  (Therefore no piece
 *		of data may be larger than 2048 bytes)  For each block to be written
 *		a long word flag is added to the header followed by ZERO long word
 *
 *		Loading would look something like this:
 *	
 *			Get length of entire file and divide by the block size
 *			to get the number of blocks
 *		
 *			for each block
 *	
 *				Load a block, record the address of the loaded block
 *				in the long word
 *	
 *			endfor
 *	
 *			//
 *			// now all the blocks are loaded and you know where the
 *			// data offset table starts so
 *			//
 *	
 *			for each non-zero data offset
 *	
 *				look up the correct block address
 *				add the block offset
 *				store in the data offset
 *	
 *			endfor
 *	
 *
 * HISTORY
 *	
 * TODO
 *
 *		* Add Best Fit Algorithm
 *
 *
 *
*/

#include "platform.h"
#include "switches.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <echidna/utils.h>
#include <echidna/eerrors.h>
#include <echidna/argparse.h>
#include <echidna/datafile.h>
#include <echidna/eio.h>
#include <echidna/strings.h>
#include <echidna/listapi.h>
#include <echidna/checkglu.h>
#include <echidna/readini.h>
#include <echidna/hash.h>

/**************************** C O N S T A N T S ***************************/

#define	PRELOAD_VERSION	0x01010101

#define MAX_LINE	256
#define	MAX_ARGS	128

#define	BUF_SIZE	(512*1024)

/******************************** T Y P E S *******************************/

typedef struct PreLoadFile
{
	LST_NODE			node;
	long				size;
	uint8*				data;
	char*				sameAs;
	struct PreLoadFile*	sameAsFile;
}
PreLoadFile;

typedef struct
{
	LST_NODE		node;
	LST_LIST		partsListX;
	LST_LIST*		partsList;
	long			numParts;
	long			size;
	struct FileContents	*fc;
}
Level;

typedef struct FileContents
{
	LST_NODE	 Node;
	uint8*		 Data;
	int			 bFromPreLoad;
	long		 Size;
	long		 Offset;
	long		 PadSize;
	struct FileContents	*SameAs;
	Level		*Level;

	uint32		 hashValue;
	BOOL		 bHasHashValue;
}
FileContents;

typedef struct
{
	LST_NODE		 Node;
	FileContents	*fc;
}
PositionNode;

#define PART_DATA	1
#define PART_LEVEL	2
#define	PART_LONG	3
#define	PART_WORD	4
#define	PART_BYTE	5
#define PART_STRING	6
#define	PART_PATH	7
#define PART_FLOAT	8

typedef struct
{
	LST_NODE		 node;
	long			 type;
	char			*string;
	FileContents	*fc;
	Level			*level;
}
Part;

/****************************** G L O B A L S *****************************/

int				 DontOut      = FALSE;
int				 Verbose      = FALSE;
int				 Pack         = TRUE;
int				 OutMode	  = 0;
int				 UseFixups    = FALSE;
int				 PadEnd		  = TRUE;
long			 PadSize	  = 4;
long			 CrossSize    = 2048;
long			 HardwareSectorSize = 1;
int				 LittleEndian = TRUE;
long			 BytesWritten = 0;
long			 EndPadBytes  = 0;
long			 HardwarePadBytes = 0;

long			 blocksMarked;

HashTable*		 FileContentsTable;
HashTable*		 PreLoadTable;
HashTable*		 FileTable;

LST_LIST		 PreLoadFileListX;
LST_LIST		*PreLoadFileList = &PreLoadFileListX;

LST_LIST		 FileListX;
LST_LIST		*FileList = &FileListX;

LST_LIST		 PosListX;
LST_LIST		*PosList = &PosListX;

LST_LIST		 SortedListX;
LST_LIST		*SortedList = &SortedListX;

LST_LIST		 LevelListX;
LST_LIST		*LevelList = &LevelListX;

long			 guessSize;
long			 TotalFiles;
long			 TotalFixups;

char			 line[MAX_LINE];
char			*args[MAX_ARGS];

char			*inputPath = "";

char			*topSection = "Start";

uint8*			 buf;

/******************************* M A C R O S ******************************/

/******************************* T A B L E S ******************************/

uint8			 ZeroData[256];

/***************************** R O U T I N E S ****************************/

/*************************************************************************
                               hashGetLong                               
 *************************************************************************

   SYNOPSIS
		char* hashGetLong (char *s, uint32 *l)

   PURPOSE
  		
  
   INPUT
		s :
		l :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

char* MK_hashGetLong (char *s, uint32 *l)
{
	uint32	v[4];
	int		i;

	for (i = 0; i < 4; i++)
	{
		v[i] = (*s) ? tolower(*s) : 0;
		s   += (*s) ? 1  : 0;
	}

	*l = ((v[3] << 24) | (v[2] << 16) | (v[1] << 8) | (v[0]));

	return s;
}


/*************************************************************************
                            LST_NodeHashFunc                             
 *************************************************************************

   SYNOPSIS
		uint32 LST_NodeHashFunc (HashEntry *he)

   PURPOSE
  		
  
   INPUT
		he :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

uint32 LST_NodeHashFunc (HashEntry *he)
{
	uint32	hash = 0;
	uint32	val  = 0;
	char	*s;

	s = LST_NodeName (he->data);
	while (*s)
	{
		s    = MK_hashGetLong (s, &val);
		hash = hash ^ val;
	}

	return hash;
}

/*************************************************************************
                           LST_NodeHashCmpFunc                           
 *************************************************************************

   SYNOPSIS
		int LST_NodeHashCmpFunc (HashEntry *t1, HashEntry *t2)

   PURPOSE
  		
  
   INPUT
		t1 :
		t2 :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int LST_NodeHashCmpFunc (HashEntry *t1, HashEntry *t2)
{
	return (stricmp (LST_NodeName (t1->data), LST_NodeName(t2->data)));
}

/*************************************************************************
                            FileContentsHashFunc                             
 *************************************************************************

   SYNOPSIS
		uint32 FileContentsHashFunc (HashEntry *he)

   PURPOSE
  		
  
   INPUT
		he :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

uint32 FileContentsHashFunc (HashEntry *he)
{
	uint32	hash = 0;
	uint32	val  = 0;
	FileContents*	fc;

	fc = (FileContents*)he->data;
	
	if (fc->bHasHashValue)
	{
		return fc->hashValue;
	}
	
	if (Pack)
	{
		uint8*	data = fc->Data;
		long	size = fc->Size - (fc->Size % 4);
	
		while (size)
		{
			hash += ((uint32)data[0]      ) +
					((uint32)data[1] <<  8) +
					((uint32)data[2] << 16) +
					((uint32)data[3] << 24) ;
			data += 4;
			size -= 4;
		}
		
		fc->hashValue     = hash;
		fc->bHasHashValue = TRUE;
	}
	else
	{
		static int bogusHashValue;
		
		fc->hashValue     = ++bogusHashValue;
		fc->bHasHashValue = TRUE;
	}

	return hash;
}

/*************************************************************************
                           FileContentsHashCmpFunc                           
 *************************************************************************

   SYNOPSIS
		int FileContentsHashCmpFunc (HashEntry *t1, HashEntry *t2)

   PURPOSE
  		
  
   INPUT
		t1 :
		t2 :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int FileContentsHashCmpFunc (HashEntry *t1, HashEntry *t2)
{
	if (((FileContents*)t1->data)->Size != ((FileContents*)t2->data)->Size)
	{
		return TRUE;
	}
	
	return (
			memcmp (
			((FileContents*)t1->data)->Data,
			((FileContents*)t2->data)->Data,
			((FileContents*)t2->data)->Size
			)
			);
}

/*************************************************************************
                             FindPreLoadFile                             
 *************************************************************************

   SYNOPSIS
		PreLoadFile* FindPreLoadFile (char* name)

   PURPOSE
  		
  
   INPUT
		name :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

PreLoadFile* FindPreLoadFile (char* name)
{
	static	PreLoadFile	findPLF;
	
	PreLoadFile*	pPLF = NULL;
	HashEntry		find;
	HashEntry*		he;
	
	LST_NodeName(&findPLF) = name;
	find.data = &findPLF;
	
	he = HASH_FindHashEntry (PreLoadTable, &find);
	
	if (he != NULL)
	{
		pPLF = (PreLoadFile*)he->data;
	}

	return pPLF;
}

/*************************************************************************
                             FindFileContentsByName                            
 *************************************************************************

   SYNOPSIS
		FileContents* FindFileContentsByName (char* name)

   PURPOSE
  		
  
   INPUT
		name :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

FileContents* FindFileContentsByName (char* name)
{
	static	FileContents	findFC;
	
	FileContents*	pFC = NULL;
	HashEntry		find;
	HashEntry*		he;
	
	LST_NodeName(&findFC) = name;
	find.data = &findFC;
	
	he = HASH_FindHashEntry (FileTable, &find);
	
	if (he != NULL)
	{
		pFC = (FileContents*)he->data;
	}

	return pFC;
}

/*************************************************************************
                        FindFileContentsByContent                        
 *************************************************************************

   SYNOPSIS
		FileContents* FindFileContentsByContent (FileContents* fc)

   PURPOSE
  		
  
   INPUT
		fc :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

FileContents* FindFileContentsByContent (FileContents* fc)
{
	FileContents*	pFC = NULL;
	HashEntry		find;
	HashEntry*		he;
	
	find.data = fc;
	he = HASH_FindHashEntry (FileContentsTable, &find);
	
	if (he != NULL)
	{
		pFC = (FileContents*)he->data;
	}

	return pFC;
}

/*************************************************************************
                            AddFileToFileList                            
 *************************************************************************

   SYNOPSIS
		void AddFileToFileList (FileContents* fc)

   PURPOSE
  		
  
   INPUT
		fc :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void AddFileToFileList (FileContents* fc)
{
	LST_AddTail (FileList, fc);
	TotalFiles++;
	
	{
		HashEntry	*he;
	
		he = CHK_CallocateMemory (sizeof (HashEntry), "Hash Entry");
		he->data = fc;
	
		HASH_AddHashEntry (FileTable, he);
	}
}


long FilePosition (FileContents *fc)
{
	return ((fc->SameAs) ? fc->SameAs->Offset : fc->Offset);
}

long MK_Write (int fh, void *data, long len)
{
	BytesWritten += len;

	return CHK_Write (fh, data, len);
}

/*********************************************************************
 *
 * WritePosition
 *
 * SYNOPSIS
 *		void  WritePosition (int fh, long position)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void WritePosition (int fh, long position, char *msg)
{

	uint32	pos;
	long	block;
	long	offset;

	block    = position / CrossSize;
	offset   = position % CrossSize;

	pos = offset | (block << 20) | 0x80000001UL;

	if (Verbose)
	{
		EL_printf ("Block $%04lx : Offset $%04lx : %s\n", block, offset, msg);
	}

	{
		if (LittleEndian)
		{
			MakeLilLong (pos);
		}
		else
		{
			MakeBigLong (pos);
		}
		MK_Write (fh, &pos,  sizeof (uint32));
	}
}
// WritePosition

/*************************************************************************
                              AddLoadedFile                              
 *************************************************************************

   SYNOPSIS
		FileContents* AddLoadedFile (char *filename, void* data, long len)

   PURPOSE
  		Add a file that's already been loaded
  
   INPUT
		filename :
		data     :
		len      :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/16/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

FileContents* AddLoadedFile (char *filename, uint8* buf, long size, int preLoad)
{
	FileContents	*newfc = NULL;

	if (size > CrossSize)
	{
		ErrMess ("File '%s' is larger than CrossSize %ld\n", filename, CrossSize);
	}
	
	if (size == 0)
	{
		WarnMess ("File '%s' is zero bytes long\n", filename);
	}

	//
	// see if it's the same as a previous file
	//
	{
		FileContents	*fc;

		newfc = CHK_CreateNode (sizeof (FileContents), filename, "FileContents");

		if (Pack)
		{
			newfc->Data = buf;
			newfc->Size = size;
			
			fc = FindFileContentsByContent (newfc);
			
			newfc->Data = NULL;
			newfc->Size = 0;
		}

		if (Pack && (fc != NULL))
		{
			//
			// it was the same
			//
			newfc->SameAs = fc;
		}
		else
		{
			//
			// it was different
			//
			PositionNode	*pos;

			pos   = CHK_CreateNode (sizeof (PositionNode), NULL, "PositionNode");

			pos->fc = newfc;
			LST_AddTail (PosList, pos);

			newfc->Size   = size;

			size  = (size + PadSize - 1) / PadSize;
			size *= PadSize;

			newfc->PadSize = size;

			guessSize += size;

			if (Pack)
			{
				newfc->Data = buf;
			}
			
			// add it to hash table
			{
				HashEntry	*he;
			
				he = CHK_CallocateMemory (sizeof (HashEntry), "Hash Entry");
				he->data = newfc;
			
				HASH_AddHashEntry (FileContentsTable, he);
			}
		}

		AddFileToFileList (newfc);
	}

	return newfc;
}

/*********************************************************************
 *
 * AddFile
 *
 * SYNOPSIS
 *		FileContents *AddFile (char *filename)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
FileContents* AddUnloadedFile (char *filename)
{
	FileContents	*newfc = NULL;
	long			 size;
	uint8			*buf = NULL;
	int				 preLoad = FALSE;

	//
	// check if this file is pre-loaded
	//
	{
		PreLoadFile*	pPLF;
		
		pPLF = FindPreLoadFile (filename);
		
		if (pPLF != NULL)
		{
			// found
			preLoad = TRUE;
			
			size = pPLF->size;
			buf  = pPLF->data;
		}
		else
		{
			// not found
			int				 fh;
			fh   = CHK_ReadOpen (filename);
			size = CHK_FileLength (fh);
		
			if (size > CrossSize)
			{
				ErrMess ("File '%s' is larger than CrossSize %ld\n", filename, CrossSize);
			}
			
			if (size == 0)
			{
				WarnMess ("File '%s' is zero bytes long\n", filename);
			}
		
			if (Pack)
			{
				if (Verbose)
				{
					EL_printf ("Reading %14s : size %6ld", filename, size);
				}
		
				buf = (uint8*)CHK_AllocateMemory (size, filename);
				CHK_Read (fh, buf, size);
			}
			CHK_Close (fh);
		}
	}

	//
	// see if it's the same as a previous file
	//
	newfc = AddLoadedFile (filename, buf, size, preLoad);
	if (newfc->SameAs != NULL)
	{
		if (Pack)
		{
			//
			// it was the same
			//
			if (!preLoad)
			{
				CHK_DeallocateMemory (buf, filename);
			}
			if (Verbose)
			{
				EL_printf (" : Same as %s", LST_NodeName(newfc->SameAs));
			}
		}
	}
		
	if (Pack && Verbose)
	{
		EL_printf ("\n");
	}

	return newfc;
}
// AddUnloadedFile

/*************************************************************************
                                 AddFile                                 
 *************************************************************************

   SYNOPSIS
		FileContents* AddFile (char* fname)

   PURPOSE
  		
  
   INPUT
		fname :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/17/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

FileContents* AddFile (char* fname)
{
	char			 filename[EIO_MAXPATH];
	FileContents*	 fc;
	
	EIO_fnmerge (filename, inputPath, fname, NULL);
	
	// check if something is loaded with the same name
	fc = FindFileContentsByName (filename);
	if (fc)
	{
		FileContents*	newfc;
		
		newfc = CHK_CreateNode (sizeof (FileContents), filename, "FileContents");
		
		// find original file
		while (fc->SameAs != NULL)
		{
			fc = fc->SameAs;
		}
		newfc->SameAs = fc;
		
		AddFileToFileList (newfc);

		return newfc;
	}
	
	return AddUnloadedFile (filename);
}

Level* FindLevel (char *levelName)
{
	char				 sectionName[256];

	sprintf (sectionName, "[%s]", levelName);

	return (Level*)LST_FindIName (LevelList, sectionName);
}
// FindLevel

/*********************************************************************
 *
 * ParseLevel
 *
 * SYNOPSIS
 *		Level *ParseLevel (IniList *specFile, char *levelName)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
Level* ParseLevel (IniList *specFile, char *levelName)
{
	char				 sectionName[256];
	SectionTracker		 secx;
	SectionTracker		*sec = &secx;
	Level				*level;

	if (Verbose)	EL_printf ("Parsing Level [%s]\n", levelName);

	sprintf (sectionName, "[%s]", levelName);

	sec = FindSection (sec, specFile, sectionName);
	if (!sec)
	{
		ErrMess ("Couldn't find Section %s\n", sectionName);
		return NULL;
	}

	level = CHK_CreateNode (sizeof (Level), sectionName, sectionName);

	level->partsList = &level->partsListX;
	LST_InitList (level->partsList);
	LST_AddTail (LevelList, level);
	
	{
		ConfigLine		*cl;

		while ((cl = GetNextINILine (sec)) != NULL)
		{
			Part	*part;

			part = CHK_CreateNode (sizeof (Part), LST_NodeName(cl), LST_NodeName (cl));
			LST_AddTail (level->partsList, part);
			level->numParts++;

			if (strlen(LST_NodeName(cl)) > MAX_LINE)
			{
				ErrMess ("Line# %d to long\n", GetConfigLineNo (cl));
			}
			else
			{
				if (!strnicmp ("Data=", LST_NodeName (cl), 5))
				{
					//
					// found a data part
					//

					part->type   = PART_DATA;
					part->fc     = AddFile (LST_NodeName (cl) + 5);
					level->size += sizeof (void *);
				}
				else if (!strnicmp ("File=", LST_NodeName (cl), 5))
				{
					//
					// found a data part
					//

					part->type   = PART_DATA;
					part->fc     = AddFile (LST_NodeName (cl) + 5);
					level->size += sizeof (void *);
				}
				else if (!strnicmp ("Level=", LST_NodeName (cl), 6))
				{
					//
					// found a level part
					//
					part->type   = PART_LEVEL;
					level->size += sizeof (void *);

					part->level = FindLevel (LST_NodeName (cl) + 6);
					if (!part->level)
					{
						part->level = ParseLevel (specFile, LST_NodeName(cl) + 6);
					}
				}
				else if (!strnicmp ("Pntr=", LST_NodeName (cl), 5))
				{
					//
					// found a level part
					//
					part->type   = PART_LEVEL;
					level->size += sizeof (void *);

					part->level = FindLevel (LST_NodeName (cl) + 5);
					if (!part->level)
					{
						part->level = ParseLevel (specFile, LST_NodeName(cl) + 5);
					}
				}
				else if (!strnicmp ("Long=", LST_NodeName(cl), 5))
				{
					int	 numargs;

					strcpy (line, LST_NodeName(cl) + 5);
					numargs = argify (line, MAX_ARGS, args);

					//
					// found a Long part
					//
					part->type   = PART_LONG;
					part->string = LST_NodeName(cl) + 5;
					level->size += sizeof (uint32) * numargs;
				}
				else if (!strnicmp ("Word=", LST_NodeName(cl), 5))
				{
					//
					// found a Word part
					//
					int	 numargs;

					strcpy (line, LST_NodeName(cl) + 5);
					numargs = argify (line, MAX_ARGS, args);

					part->type   = PART_WORD;
					part->string = LST_NodeName(cl) + 5;
					level->size += sizeof (uint16) * numargs;
				}
				else if (!strnicmp ("Byte=", LST_NodeName(cl), 5))
				{
					//
					// found a Byte part
					//
					int	 numargs;

					strcpy (line, LST_NodeName(cl) + 5);
					numargs = argify (line, MAX_ARGS, args);

					part->type   = PART_BYTE;
					part->string = LST_NodeName(cl) + 5;
					level->size += sizeof (uint8) * numargs;
				}
				else if (!strnicmp ("Float=", LST_NodeName(cl), 6))
				{
					//
					// found a Word part
					//
					int	 numargs;

					strcpy (line, LST_NodeName(cl) + 6);
					numargs = argify (line, MAX_ARGS, args);

					part->type   = PART_FLOAT;
					part->string = LST_NodeName(cl) + 6;
					level->size += sizeof (float) * numargs;
				}
				else if (!strnicmp ("String=", LST_NodeName(cl), 7))
				{
					//
					// found a String part
					//
					part->type   = PART_STRING;
					part->string = LST_NodeName(cl) + 7;
					level->size += strlen (LST_NodeName(cl) + 7);
				}
				else if (!strnicmp ("Path=", LST_NodeName(cl), 7))
				{
					//
					// found a Path part
					//
					part->type   = PART_PATH;
					part->string = LST_NodeName(cl) + 5;
				//	level->size += 0;
					inputPath = part->string;
				}
				else
				{
					ErrMess ("Unknown specFile line #%d:'%s'\n", GetConfigLineNo (cl), LST_NodeName (cl));
				}
			}
		}
	}

	return level;
}
// ParseLevel

/*************************************************************************
                               ReadPreLoad                               
 *************************************************************************

   SYNOPSIS
		int ReadPreLoad (char *filename)

   PURPOSE
  		
  
   INPUT
		filename :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   RETURNS
  
  
   SEE ALSO
  
  
   HISTORY
		01/16/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int ReadPreLoad (char *filename)
{
	int			done = FALSE;
	int			fh;
	long		len;
	uint8*		data;
	LST_LIST	sameList;
	
	LST_InitList (&sameList);
	
	// read entire file into memory	
	fh   = CHK_ReadOpen (filename);
	len  = CHK_FileLength (fh);
	data = (uint8*)CHK_AllocateMemory (len, filename);
	
	CHK_Read (fh, data, len);
	CHK_Close (fh);
	
	// check version
	{
		long	version;
		
		version = *((long*)data);
		data   += 4;
		version = LSBFToNative32Bit(version);
		
		if (version != PRELOAD_VERSION)
		{
			FailMess ("Bad version for pre load file %s\n", filename);
		}
	}
	
	// add files to PreLoadFilelist
	while (!done)
	{
		PreLoadFile*	pPLF;
		char*	name;
		long	size;
		long	type;
		
		type = *((long*)data);
		data += 4;
		type  = LSBFToNative32Bit(type);
		
		switch (type)
		{
		case 0:	// end of file
			done = TRUE;
			break;
		case 1: // file data
			{
				size  = *((long*)data);
				data += 4;
				name  = data;
				data += strlen (name) + 1;
				size  = LSBFToNative32Bit(size);
				
				if (Verbose)
				{
					EL_printf ("Preloading %14s : size %6ld\n", name, size);
				}
				
				pPLF = (PreLoadFile*)CHK_CreateNode (sizeof (PreLoadFile), name, "PreLoadFile");
				
				pPLF->size = size;
				pPLF->data = data;
				
				data += size;
				
				LST_AddTail (PreLoadFileList, pPLF);
			
				{
					HashEntry	*he;
				
					he = CHK_CallocateMemory (sizeof (HashEntry), "Hash Entry");
					he->data = pPLF;
				
					HASH_AddHashEntry (PreLoadTable, he);
				}
			}
			break;
		case 2:	// same as
			{
				char*	sameName;
				
				name      = data;
				data     += strlen (name) + 1;
				sameName  = data;
				data     += strlen (sameName) + 1;
				
				if (Verbose)
				{
					EL_printf ("Preloading %14s : size %6ld : same as %14s\n", name, size, sameName);
				}
				
				pPLF = (PreLoadFile*)CHK_CreateNode (sizeof (PreLoadFile), name, "PreLoadFile");
				
				pPLF->sameAs = sameName;
				
				LST_AddTail (&sameList, pPLF);
			}
			break;
		default:
			FailMess ("unknown type %d in preload file %s\n", type, filename);
			break;
		}
	}
	
	// add same as ones to list now that we're sure the ones that they're the same
	// as ones that are loaded
	{
		PreLoadFile* pPLF;
		
		while ((pPLF = (PreLoadFile*)LST_RemHead (&sameList)) != NULL)
		{
			PreLoadFile*	samePLF;
			HashEntry*		he;
			
			samePLF = FindPreLoadFile (pPLF->sameAs);
			
			ENSURE_F (samePLF == NULL, ("Impossible! Didn't find a match\n"));
			
			pPLF->data       = samePLF->data;
			pPLF->size       = samePLF->size;
			pPLF->sameAsFile = samePLF;
			
			LST_AddTail (PreLoadFileList, pPLF);
			
			he = CHK_CallocateMemory (sizeof (HashEntry), "Hash Entry");
			he->data = pPLF;
			
			HASH_AddHashEntry (PreLoadTable, he);
		}
	}
	
	RETURN TRUE;
}

/*************************************************************************
                              WritePreLoad                               
 *************************************************************************

   SYNOPSIS
		void WritePreLoad (char* filename)

   PURPOSE
  		
  
   INPUT
		filename :
  
   OUTPUT
		None  
  
   EFFECTS
		None  
  
   SEE ALSO
  
  
   HISTORY
		01/16/97 GAT: Created.
  
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void WritePreLoad (char* filename, char* keyname)
{
	long zero = NativeToLSBF32Bit(0);
	long one  = NativeToLSBF32Bit(1);
	long two  = NativeToLSBF32Bit(2);

	if (filename)
	{
		FileContents*	fc;
		int	fh;
		
		fh = CHK_WriteOpen (filename);
		
		{
			long	version = NativeToLSBF32Bit(PRELOAD_VERSION);
			
			CHK_Write (fh, &version, sizeof (version));
		}
	
		fc = (FileContents*)LST_Head (FileList);
		while (!LST_EndOfList (fc))
		{
			if (!fc->bFromPreLoad)
			{
				FileContents*	dataFc;
				
				dataFc = fc;
				while (dataFc->SameAs != NULL)
				{
					dataFc = dataFc->SameAs;
				}
				
				if (dataFc->Data)
				{
					if (fc->SameAs != NULL)
					{
						long	size;
						
						size = fc->Size;
						size = NativeToLSBF32Bit(size);
						
						CHK_Write (fh, &two, sizeof (two));
						CHK_Write (fh, LST_NodeName(fc), strlen (LST_NodeName(fc)) + 1);
						CHK_Write (fh, LST_NodeName(dataFc), strlen (LST_NodeName(dataFc)) + 1);
					}
					else
					{
						long	size;
						
						size = fc->Size;
						size = NativeToLSBF32Bit(size);
						
						CHK_Write (fh, &one, sizeof (one));
						CHK_Write (fh, &size, sizeof (size));
						CHK_Write (fh, LST_NodeName(fc), strlen (LST_NodeName(fc)) + 1);
						if (fc->Size)
						{
							CHK_Write (fh, fc->Data, fc->Size);
					    }
					}
				}
			}
			fc = (FileContents*)LST_Next (fc);
		}
	
		CHK_Write (fh, &zero, sizeof (zero));
		
		CHK_Close (fh);
	}
	
	if (keyname)
	{
		LST_LIST		SortList;
		FileContents*	fc;
		
		LST_InitList (&SortList);
		
		fc = (FileContents*)LST_Head (FileList);
		while (!LST_EndOfList (fc))
		{
			LST_NODE*	nd;
			
			nd = LST_Head (&SortList);
			while (!LST_EndOfList(nd))
			{
				int	r;

				r = stricmp (LST_NodeName(fc), LST_NodeName(nd));
				if (r < 0)
				{
					LST_InsertBefore (nd, CHK_CreateNode(sizeof(LST_NODE), LST_NodeName(fc), "SortNode"));
					break;
				}
				if (r == 0)
				{
					break;
				}
				nd = LST_Next (nd);
			}
			
			if (LST_EndOfList(nd))
			{
				LST_InsertBefore (nd, CHK_CreateNode(sizeof(LST_NODE), LST_NodeName(fc), "SortNode"));
			}
			
			fc = (FileContents*)LST_Next(fc);
		}
		
		{
			LST_NODE*	nd;
			FILE*		fp;

			fp = CHK_fopen (keyname, "w");
		
			nd = LST_Head (&SortList);
			while (!LST_EndOfList(nd))
			{
				strlwr (LST_NodeName(nd));
				CHK_fprintf (fp, "%s\n", LST_NodeName(nd));

				nd = LST_Next (nd);
			}
		
			CHK_fclose(fp);
		}
	}
}

/******************************** TEMPLATE ********************************/

#define ARG_OUTFILE		(newargs[ 0])
#define ARG_SPECFILE	(newargs[ 1])
#define ARG_VERBOSE		(newargs[ 2])
#define	ARG_DONTOUT		(newargs[ 3]) 
#define ARG_PACK		(newargs[ 4]) 
#define ARG_BIGENDIAN	(newargs[ 5]) 
#define ARG_BYTES		(newargs[ 6]) 
#define ARG_CROSS		(newargs[ 7]) 
#define	ARG_HARDWARE	(newargs[ 8]) 
#define ARG_FIXUPS		(newargs[ 9]) 
#define	ARG_PADEND		(newargs[10]) 
#define ARG_TOPSECTION	(newargs[11]) 
#define	ARG_WRITELOAD	(newargs[12]) 
#define	ARG_KEYFILE		(newargs[13]) 
#define	ARG_READLOAD	(newargs[14])

char Usage[] = "Usage: MKLOADOB OUTFILE SPECFILES [switches...]\n";

ArgSpec Template[] = {
{STANDARD_ARG|REQUIRED_ARG,				"OUTFILE",	"\tOUTFILE     = Output filename\n", },
{STANDARD_ARG|REQUIRED_ARG|MULTI_ARG,	"SPECFILES","\tSPECFILES   = File(s) of levels and data\n", },
{CHRSWITCH_ARG,							"V",		"\t-V          = Verbose (show errors)\n", },
{CHRSWITCH_ARG,							"D",		"\t-D          = Don't write output file\n", },
{CHRSWITCH_ARG,							"P",		"\t-P          = Don't Pack\n", },
{CHRSWITCH_ARG,							"B",		"\t-B          = Big Endian\n", },
{CHRKEYWORD_ARG,						"S",		"\t-S<bytes>   = Sector Size (Def. 4)\n", },
{CHRKEYWORD_ARG,						"C",		"\t-C<bytes>   = Cross Size (Def. 2048)\n", },
{CHRKEYWORD_ARG,						"A",		"\t-A<bytes>   = Hardware Sector Size (Def. 1)\n", },
{CHRSWITCH_ARG,							"F",		"\t-F          = Put Fixups in file\n", },
{CHRSWITCH_ARG,							"E",		"\t-E          = Don't pad end of file to cross size\n", },
{CHRKEYWORD_ARG,						"T",		"\t-T<top>     = Top Section (Def. 'START')\n", },
{CHRKEYWORD_ARG,						"W",		"\t-W<prefile> = Write Preload file\n", },
{CHRKEYWORD_ARG,						"K",		"\t-K<prekey>  = Write Preload key file\n", },
{CHRKEYWORD_ARG|MULTI_ARG,				"R",		"\t-R<prefile> = Read prefile\n", },
{0, NULL, NULL, },
};

/********************************** MAIN **********************************/

int main(int argc, char **argv)
{
	char	**newargs;
	long	 blockTableSize;
	long	 offsetTableSize;
	long	 totalSize;

	EL_printf ("MKLOADOB Copyright (c) 1997 Echidna\n");

	LST_InitList (PreLoadFileList);
	LST_InitList (FileList);
	LST_InitList (PosList);
	LST_InitList (SortedList);
	LST_InitList (LevelList);
	
	PreLoadTable = HASH_CreateHashTable (
							LST_NodeHashFunc,
							LST_NodeHashCmpFunc,
							4096);
								
	FileTable = HASH_CreateHashTable (
							LST_NodeHashFunc,
							LST_NodeHashCmpFunc,
							4096);
	
	FileContentsTable = HASH_CreateHashTable (
							FileContentsHashFunc,
							FileContentsHashCmpFunc,
							4096);
	

	SetINIMergeSections(FALSE);
	SetINICaseSensitive(FALSE);

	newargs = argparse (argc, argv, Template);

	if (!newargs)
	{
		EL_printf ("%s\n", GlobalErrMsg);
		printarghelp (Usage, Template);
		return EXIT_FAILURE;
	}
	else
	{

		Verbose      =  SWITCH_VALUE(ARG_VERBOSE);
		Pack         = !SWITCH_VALUE(ARG_PACK);
		LittleEndian = !SWITCH_VALUE(ARG_BIGENDIAN);
		UseFixups    =  SWITCH_VALUE(ARG_FIXUPS);
		PadEnd       = !SWITCH_VALUE(ARG_PADEND);
		DontOut      =  SWITCH_VALUE(ARG_DONTOUT);

		if (ARG_TOPSECTION)
		{
			topSection = ARG_TOPSECTION;
		}

		if (ARG_BYTES)		PadSize            = EL_atol (ARG_BYTES);
		if (ARG_CROSS)		CrossSize          = EL_atol (ARG_CROSS);
		if (ARG_HARDWARE)	HardwareSectorSize = EL_atol (ARG_HARDWARE);
		
		if (PadSize <= 1)
		{
			FailMess ("Sector Size must be > 1\n");
		}
		
		buf = CHK_AllocateMemory (BUF_SIZE, "copy buffer");

		//
		// load pre-load stuff
		//
		{
			char			**files;
			
			files = MULTI_ARGLIST (ARG_READLOAD);
			if (files)
			{
				while (*files)
				{
					if (Verbose)	EL_printf ("=====Reading Preload File %s=====\n", *files);

					ReadPreLoad (*files);

					files++;
				}

			}
		}

		//---------------------------------------------
		// Parse SPECFILE and create list of data files to load and levels
		//---------------------------------------------
		{
			IniList			*specFile = NULL;
			char			**files;

			files = MULTI_ARGLIST (ARG_SPECFILE);

			while (*files)
			{
				if (Verbose)	EL_printf ("=====Reading Spec File %s=====\n", *files);

				specFile = AppendINI (specFile, *files);
				if (!specFile)
				{
					FailMess ("Reading specfile '%s'\n", *files);
				}
				files++;
			}
			ParseLevel (specFile, topSection);
		}

		//---------------------------------------------
		// add level 'parts'
		//---------------------------------------------
		if (!ErrorCount)
		{
			Level	*level;

			level = (Level*)LST_Head (LevelList);

			if (Verbose)	EL_printf ("=====Adding Level Parts=====\n");

			while (!LST_EndOfList (level))
			{

				long			 size;
				
				size = level->size;
				
				if (Verbose)	EL_printf ("=====Adding Level %s=====\n", LST_NodeName (level));

				if (size > CrossSize)
				{
					ErrMess ("File '%s' is larger than CrossSize %ld\n", LST_NodeName(level), CrossSize);
				}
				
				if (size == 0)
				{
					WarnMess ("Level '%s' is zero bytes long\n", LST_NodeName(level));
				}

				{
					FileContents	*newfc;
					PositionNode	*pos;

					newfc = CHK_CreateNode (sizeof (FileContents), LST_NodeName(level), "FileContents");
					newfc->Level = level;
					level->fc    = newfc;

					pos   = CHK_CreateNode (sizeof (PositionNode), NULL, "PositionNode");

					pos->fc = newfc;
					LST_AddTail (PosList, pos);

					newfc->Size   = size;

					size  = (size + PadSize - 1) / PadSize;
					size *= PadSize;

					newfc->PadSize = size;

					guessSize += size;

					AddFileToFileList (newfc);
				}

				level = (Level*)LST_Next (level);
			}
		}

		if (ARG_WRITELOAD || ARG_KEYFILE)
		{
			WritePreLoad (ARG_WRITELOAD, ARG_KEYFILE);
		}

		if (DontOut)
		{
			goto done;
		}

/************************* Sort and Position files ************************/

		if (!ErrorCount)
		{
			long	count = 0;
			long	curAddress;

			offsetTableSize = (UseFixups ? ((TotalFixups + 1) * sizeof (uint32)) : 0) + sizeof (uint32);

			//
			// try until it works
			//
			for (;;)
			{
				if (Verbose)
				{
					EL_printf ("------------------------\n");
					EL_printf ("starting guessSize = %ld\n", guessSize);
				}

				blockTableSize  = (((guessSize + CrossSize - 1) / CrossSize) + 1) * sizeof (uint32);
				curAddress      = blockTableSize + offsetTableSize;
				guessSize		= guessSize + blockTableSize + offsetTableSize;

				if (Verbose)
				{
					EL_printf ("new guessSize  = %ld\n", guessSize);
					EL_printf ("blockTableSize = %ld\n", blockTableSize);
				}

				//
				// sort files largest to smallest
				//
				{
					PositionNode	*pos;

					while ((pos = (PositionNode*)LST_RemHead (PosList)) != NULL)
					{
						PositionNode	*old;

						old = (PositionNode*)LST_Head (SortedList);
						for (;;)
						{
							if (LST_EndOfList (old))
							{
								LST_InsertBefore (old, pos);
								break;
							}
							if (pos->fc->Size > old->fc->Size)
							{
								LST_InsertBefore (old, pos);
								break;
							}
							old = (PositionNode*)LST_Next (old);
						}
					}
				}

				//
				// add files to list
				//
				while (!LST_IsEmpty (SortedList))
				{
					PositionNode	*pos;
					FileContents	*fc;

					pos = (PositionNode*)LST_Head (SortedList);
					while (!LST_EndOfList (pos))
					{
						long			 endAddress;

						fc = pos->fc;

						endAddress = curAddress + fc->PadSize;

						if ((curAddress + CrossSize - 1) / CrossSize == (endAddress + CrossSize - 1) / CrossSize)
						{
							break;
						}

						pos = (PositionNode*)LST_Next (pos);
					}

					if (LST_EndOfList (pos))
					{
						//
						// No file that fits was found so look move to next
						// boundry and just use first file
						//
						curAddress  = (curAddress + CrossSize - 1) / CrossSize;
						curAddress *= CrossSize;

						pos = (PositionNode*)LST_Head (SortedList);
						fc  = pos->fc;
					}

					LST_Remove (pos);
					LST_AddTail (PosList, pos);

					fc->Offset  = curAddress;
					curAddress += fc->PadSize;
				}

				//
				// all the files have been added
				// did the sector count match?
				//

				if (((uint32)curAddress + CrossSize - 1) / CrossSize <=
					(blockTableSize - 1) / sizeof (uint32))
				{
					break;
				}

				count++;

				if (Verbose)
				{
					EL_printf ("Sorting %d\n", count);
				}

				guessSize = curAddress;
			}

			totalSize = curAddress;
		}

/******************************* Write Files ******************************/

		if (!ErrorCount)
		{
			int				 fh;
			long			 headerTableSize;
			long			 slack = 0;

			fh = CHK_WriteOpen (ARG_OUTFILE);

			headerTableSize = offsetTableSize + blockTableSize;

			if (headerTableSize > CrossSize)
			{
				FailMess ("To many files or size to many sectors\n");
			}

			if (Verbose)
			{
				EL_printf ("Offset Table Size = %ld\n", offsetTableSize);
				EL_printf ("block Table Size  = %ld\n", blockTableSize);
				EL_printf ("header Table Size = %ld\n", headerTableSize);

				EL_printf ("Writing %ld bytes to file %s\n", ((totalSize + CrossSize - 1) / CrossSize) * CrossSize, ARG_OUTFILE);
			}

			//---------------------------------------------
			// write block flags
			//---------------------------------------------
			{
				long	blocks;

				blocks = (totalSize + CrossSize - 1) / CrossSize;
				blocksMarked = blocks;

				while (blocks)
				{
					uint32	flag = 0xFFFFFFFFUL;

					MK_Write (fh, &flag, sizeof (uint32));

					blocks--;
				}

				{
					uint32	zero = 0;

					MK_Write (fh, &zero, sizeof (uint32));
				}
			}

			//---------------------------------------------
			// write fixups
			//---------------------------------------------
			if (UseFixups)
			{
				Level	*level;

				level = (Level*)LST_Head (LevelList);
				while (!LST_EndOfList (level))
				{
					Part	*part;
					long	 position;

					position = FilePosition (level->fc);

					part = (Part*)LST_Head (level->partsList);
					while (!LST_EndOfList (part))
					{
						switch (part->type)
						{
						case PART_DATA:
						case PART_LEVEL:
							WritePosition (fh, position, LST_NodeName(part));
							position += sizeof (uint32);
							break;
						default:
							break;
						}
						part = (Part*)LST_Next (part);
					}
					level = (Level*)LST_Next (level);
				}

				{
					uint32	zero = 0;

					MK_Write (fh, &zero, sizeof (uint32));
				}
			}

			//---------------------------------------------
			// write pointer to top level
			//---------------------------------------------
			{
				Level	*level;

				level = (Level*)LST_Head (LevelList);

				WritePosition (fh, FilePosition(level->fc), LST_NodeName(level));
			}

			#if 0
			//---------------------------------------------
			// write file offsets
			//---------------------------------------------
			{
				FileContents	*fc;

				fc = (FileContents*)LST_Head (FileList);
				while (!LST_EndOfList (fc))
				{
					WritePosition (fh, FilePosition(fc), LST_NodeName(fc));
			
					fc = (FileContents*)LST_Next (fc);
				}

				{
					uint32	zero = 0;

					MK_Write (fh, &zero, sizeof (uint32));
				}
			}
			#endif

			//---------------------------------------------
			// write files
			//---------------------------------------------
			{
				PositionNode	*pos;

				pos = (PositionNode*)LST_Head (PosList);
				while (!LST_EndOfList (pos))
				{
					FileContents	*fc;
					long			 offset;

					fc     = pos->fc;
					offset = FilePosition(fc);

					if (offset != BytesWritten)
					{
						ErrMess ("Offset $%08lx : BytesWritten $%08lx\n", offset, BytesWritten);
					}

					if (fc->Level)
					{
						if (Verbose)
						{
							EL_printf ("Writing %7ld bytes from %s\n", fc->Size, LST_NodeName (fc));
						}

						{
							Part	*part;

							part = (Part*)LST_Head (fc->Level->partsList);
							while (!LST_EndOfList (part))
							{
								switch (part->type)
								{
								case PART_DATA:
									WritePosition (fh, FilePosition(part->fc), LST_NodeName(part));
									break;
								case PART_LEVEL:
									WritePosition (fh, FilePosition(part->level->fc), LST_NodeName(part));
									break;
								case PART_LONG:
									{
										uint32	 vl;
										int	 numargs;
										int	 i;

										strcpy (line, part->string);
										numargs = argify (line, MAX_ARGS, args);

										for (i = 0; i < numargs; i++)
										{
											vl = EL_atol (args[i]);

											if (LittleEndian)
											{
												MakeLilLong(vl);
											}
											else
											{
												MakeBigLong(vl);
											}

											MK_Write (fh, &vl, sizeof (uint32));
										}
									}
									break;
								case PART_WORD:
									{
										uint16	 vw;
										int	 numargs;
										int	 i;

										strcpy (line, part->string);
										numargs = argify (line, MAX_ARGS, args);

										for (i = 0; i < numargs; i++)
										{
											vw = (uint16)EL_atol (args[i]);

											if (LittleEndian)
											{
												MakeLilWord(vw);
											}
											else
											{
												MakeBigWord(vw);
											}

											MK_Write (fh, &vw, sizeof (uint16));
										}
									}
									break;
								case PART_BYTE:
									{
										uint8	 vb;
										int	 numargs;
										int	 i;

										strcpy (line, part->string);
										numargs = argify (line, MAX_ARGS, args);

										for (i = 0; i < numargs; i++)
										{
											vb = (uint8)EL_atol (args[i]);

											MK_Write (fh, &vb, sizeof (uint8));
										}
									}
									break;
								case PART_FLOAT:
									{
										float	 vf;
										int	 numargs;
										int	 i;

										strcpy (line, part->string);
										numargs = argify (line, MAX_ARGS, args);

										for (i = 0; i < numargs; i++)
										{
											vf = (float)atof (args[i]);

											if (LittleEndian)
											{
												#if __LITTLEENDIAN__
													MK_Write (fh, &vf, sizeof (float));
												#else
													{
														char	*s;
														char	 d[4];

														s = (char *)&vf;

														d[0] = s[3];
														d[1] = s[2];
														d[2] = s[1];
														d[3] = s[0];

														MK_Write (fh, &d, sizeof (d));
													}
												#endif
											}
											else
											{
												#if __LITTLEENDIAN__
													{
														char	*s;
														char	 d[4];

														s = (char *)&vf;

														d[0] = s[3];
														d[1] = s[2];
														d[2] = s[1];
														d[3] = s[0];

														MK_Write (fh, &d, sizeof (d));
													}
												#else
													MK_Write (fh, &vf, sizeof (float));
												#endif
											}
										}
									}
									break;
								case PART_STRING:
									{
										MK_Write (fh, part->string, strlen (part->string));
									}
									break;
								}

								part = (Part*)LST_Next (part);
							}
						}
					}
					else if (fc->Size)
					{
						if (Verbose)
						{
							EL_printf ("Writing %7ld bytes from %s\n", fc->Size, LST_NodeName (fc));
						}

						if (Pack)
						{
							uint8	*data;
	
							data = fc->Data;
							MK_Write (fh, data, fc->Size);
						}
						else
						{
							int		infh;
							long	bytes;
						
							infh = CHK_ReadOpen (LST_NodeName(fc));
							if (infh == (-1))
							{
								FailMess ("Couldn't open %s\n", LST_NodeName (fc));
							}
						
							bytes = fc->Size;
							while (bytes)
							{
								long	len;
							
								len = min (bytes, BUF_SIZE);
								if (len != CHK_Read (infh, buf, len))
								{
									FailMess ("Trouble reading %s\n", LST_NodeName (fc));
								}
								if (len != MK_Write (fh, buf, len))
								{
									FailMess ("Trouble writing %s\n", ARG_OUTFILE);
								}
							
								bytes -= len;
							}

							CHK_Close (infh);
						}
					}
				
					{
						long	pad;

						pad = fc->PadSize - fc->Size;
						while (pad)
						{
							long	part;

							part = min (pad, 256);
							MK_Write (fh, ZeroData, part);
							pad -= part;
						}
					}

					pos = (PositionNode*)LST_Next (pos);

					if ((LST_EndOfList (pos) && PadEnd) || (!LST_EndOfList (pos) && FilePosition(pos->fc) / CrossSize != offset / CrossSize))
					{
						//
						// pad to next sector
						//
						long	pad;

						if (offset + fc->PadSize != BytesWritten)
						{
							ErrMess ("NOffset $%08lx : BytesWritten $%08lx\n", offset + fc->PadSize, BytesWritten);
						}

						pad = ((offset + fc->PadSize) % CrossSize);

						if (pad)
						{
							pad = CrossSize - pad;
							slack += pad;

							while (pad)
							{
								long	part;

								part = min (pad, 256);
								MK_Write (fh, ZeroData, part);
								pad -= part;
							}
						}

					}
					else if (LST_EndOfList (pos))
					{
						if (!PadEnd)
						{
							long	pad;
	
							pad = ((offset + fc->PadSize) % CrossSize);
							if (pad)
							{
								EndPadBytes = CrossSize - pad;
							}
						}
					}

					// pad to hardware sector					
					if (LST_EndOfList (pos))
					{
						long	pad;

						pad = BytesWritten % HardwareSectorSize;

						if (pad)
						{
							pad = HardwareSectorSize - pad;
							slack += pad;
							
							HardwarePadBytes += pad;

							while (pad)
							{
								long	part;

								part = min (pad, 256);
								MK_Write (fh, ZeroData, part);
								pad -= part;
							}
						}
					}
				}
			}

			if (blocksMarked * CrossSize - EndPadBytes + HardwarePadBytes != BytesWritten)
			{
				ErrMess ("Marked %ld (%ld) blocks, wrote %ld bytes\n",
					blocksMarked,
					blocksMarked * CrossSize,
					BytesWritten
				);
			}
			EL_printf ("%-30s : %5ld bytes : %3ld blocks : %5ld Slack bytes\n",
 				ARG_OUTFILE, BytesWritten, blocksMarked, slack);

			CHK_Close (fh);
		}
/************************************  ************************************/
	}

done:
	if (GlobalErr) {
		EL_printf ("%s\n", GlobalErrMsg);
		return EXIT_FAILURE;
	}

	if (ErrorCount)
	{
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}


